#!/usr/bin/env python
# -*- coding: utf-8 -*-# 
# @(#)migration_tools.py
# 
# 
# Copyright (C) 2014, GC3, University of Zurich. All rights reserved.
# 
# 
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

__docformat__ = 'reStructuredText'
__author__ = 'Antonio Messina <antonio.s.messina@gmail.com>'

import os
import sys

from elasticluster.subcommands import AbstractCommand

from elasticluster import repository, cluster

# commit 73979e27b3cd8e62d2f023d915ffd162d36bb301
# Author: Antonio Messina <antonio.s.messina@gmail.com>
# Date:   Sun Nov 2 10:56:14 2014 +0100
#
#     Rename `ClusterRepository` to `PickleRepository`
repository.ClusterRepository = repository.PickleRepository

class NotPresent:
    def __str__(self):
        return "Not present before"

def __setstate_upgrade__(self, state):
    self.__dict__ = state
    self._patches = {}

    # commit cc446ac19907ef692ec08336f06b5e6b740d6387
    # Author: Antonio Messina <antonio.s.messina@gmail.com>
    # Date:   Sat Oct 25 16:57:58 2014 +0200
    #
    #     Save host keys in `storage_path` as <clustername>.known_hosts and use it when connecting to the cluster
    old_known_hosts_file = state.get('known_hosts_file', NotPresent())

    path = "%s/%s.known_hosts" % (self.repository.storage_path, self.name)
    known_hosts_file = path if os.path.isfile(path) else None

    if old_known_hosts_file != known_hosts_file:
        self._patches['known_hosts_file'] = (NotPresent(), known_hosts_file)

    # commit 6413374799e945953492e699cca4bcabf572b5df
    # Author: Antonio Messina <antonio.s.messina@gmail.com>
    # Date:   Mon Nov 3 09:49:32 2014 +0100
    #
    #     Also save the template name of a cluster, if known.
    if 'template' not in state:
        self._patches['template'] = (NotPresent(), self.extra['template'])

    # commit 7b4ed108a699c6801b2d22790ae30368f416c246
    # Author: Antonio Messina <antonio.s.messina@gmail.com>
    # Date:   Tue Nov 4 11:36:40 2014 +0100

    #     Add a configuration option `thread_pool_max_size` to limit
    #     the maximum amount of processes that are created when
    #     starting virtual machine
    if 'thread_pool_max_size' not in state:
        self._patches['thread_pool_max_size'] = (NotPresent(), 10)

    for attr, values in self._patches.items():
        self.__dict__[attr] = values[1]

def patch_cluster():
    """
    Patch Cluster class to allow Pickle to load old clusters
    """    
    cluster.Cluster.__setstate__ = __setstate_upgrade__
    

class MigrationCommand(AbstractCommand):
    """
    Migrate stored clusters from older versions of elasticluster.
    """

    def setup(self, subparsers):
        parser = subparsers.add_parser(
            "migrate", help="Migrate a stored cluster", description=self.__doc__)
        parser.set_defaults(func=self)
        parser.add_argument('cluster', nargs='*',
                            help='Only migrate CLUSTER. By default, all clusters are migrated')
        parser.add_argument('-s', '--storage-path', metavar='PATH',
                            default=os.path.expanduser('~/.elasticluster/storage'),
                            help="Path to elasticluster storage directory. Default: %(default)s")
        parser.add_argument('-n', '--dry-run', action='store_true',
                            help="Do not actually migrate anything, just show what's going to happen")

    def execute(self):
        """
        migrate storage
        """
        repo = repository.PickleRepository(self.params.storage_path)

        clusters = [i[:-7] for i in os.listdir(self.params.storage_path) if i.endswith('.pickle')]

        if self.params.cluster:
            clusters = filter(lambda x: x in self.params.cluster, clusters)

        if not clusters:
            print("No clusters")
            sys.exit(0)
        
        patch_cluster()
        for cluster in clusters:
            print("Cluster `%s`" % cluster)
            print("path: %s" % repo.storage_path + '/%s.pickle' % cluster)
            cl = repo.get(cluster)
            if cl._patches:
                print("Attributes changed: ")
                for attr, val in cl._patches.items():
                    print("  %s: %s -> %s" % (attr, val[0], val[1]))
            else:
                print("No upgrade needed")

            print("")
            if not self.params.dry_run:
                if cl._patches:
                    del cl._patches
                    print("Changes saved to disk")
                    cl.repository.save_or_update(cl)
            
                



#!/usr/bin/env python
# -*- coding: utf-8 -*-# 
# @(#)openstack.py
# 
# 
# Copyright (C) 2013, GC3, University of Zurich. All rights reserved.
# 
# 
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

__docformat__ = 'reStructuredText'
__author__ = 'Antonio Messina <antonio.s.messina@gmail.com>'

# System imports
import os
import threading

# External modules
from novaclient import client
from novaclient.exceptions import NotFound
from paramiko import DSSKey, RSAKey, PasswordRequiredException
from paramiko.ssh_exception import SSHException

# Elasticluster imports
from elasticluster import log
from elasticluster.memoize import memoize
from elasticluster.providers import AbstractCloudProvider
from elasticluster.exceptions import SecurityGroupError, KeypairError, \
    ImageError, InstanceError, ClusterError

DEFAULT_OS_NOVA_API_VERSION = "1.1"

class OpenStackCloudProvider(AbstractCloudProvider):
    """
    This implementation of
    :py:class:`elasticluster.providers.AbstractCloudProvider` uses the
    OpenStack native python bindings to connect to OpenStack clouds
    and manage instances.

    :param str username: username of the keystone user
    :param str password: password of the keystone user
    :param str project_name: name of the project to use
    :param str auth_url: url of keystone endpoint
    :param str region: OpenStack region to use
    :param str storage_path: path to store temporary data
    :param bool request_floating_ip: Whether ip are assigned automatically
                                    `True` or floating ips have to be
                                    assigned manually `False`
    
    """
    __node_start_lock = threading.Lock()  # lock used for node startup

    def __init__(self, username, password, project_name, auth_url,
                 region_name=None, storage_path=None,
                 request_floating_ip=False,
                 nova_api_version=DEFAULT_OS_NOVA_API_VERSION):
        self._os_auth_url = os.getenv('OS_AUTH_URL', auth_url)
        self._os_username = os.getenv('OS_USERNAME', username)
        self._os_password = os.getenv('OS_PASSWORD', password)
        self._os_tenant_name = os.getenv('OS_TENANT_NAME', project_name)
        self._os_region_name = region_name
        self.request_floating_ip = request_floating_ip
        self.nova_api_version = nova_api_version
        self._instances = {}
        self._cached_instances = []

        self.client = client.Client(self.nova_api_version,
                                    self._os_username, self._os_password, self._os_tenant_name,
                                    self._os_auth_url, region_name=self._os_region_name)
 
    def start_instance(self, key_name, public_key_path, private_key_path,
                       security_group, flavor, image_id, image_userdata,
                       username=None, node_name=None, **kwargs):
        """Starts a new instance on the cloud using the given properties.
        The following tasks are done to start an instance:

        * establish a connection to the cloud web service
        * check ssh keypair and upload it if it does not yet exist. This is
          a locked process, since this function might be called in multiple
          threads and we only want the key to be stored once.
        * check if the security group exists
        * run the instance with the given properties

        :param str key_name: name of the ssh key to connect
        :param str public_key_path: path to ssh public key
        :param str private_key_path: path to ssh private key
        :param str security_group: firewall rule definition to apply on the
                                   instance
        :param str flavor: machine type to use for the instance
        :param str image_id: image type (os) to use for the instance
        :param str image_userdata: command to execute after startup
        :param str username: username for the given ssh key, default None

        :return: str - instance id of the started instance
        """

        log.debug("Checking keypair `%s`.", key_name)
        with OpenStackCloudProvider.__node_start_lock:
            self._check_keypair(key_name, public_key_path, private_key_path)

        log.debug("Checking security group `%s`.", security_group)
        self._check_security_group(security_group)

        # Check if the image id is present.
        images = self._get_images()
        if image_id not in [img.id for img in images]:
            raise ImageError("No image found with id '%s' on cloud "
                             "%s" % (image_id, self._os_auth_url))

        # Check if the flavor exists
        flavors = [fl for fl in self._get_flavors() if fl.name == flavor]
        if not flavors:
            raise FlavorError("No flavor found with name %s on cloud "
                              "%s" % (flavor, self._os_auth_url))
        flavor = flavors[0]

        nics = None
        if 'network_ids' in kwargs:
            nics=[{'net-id': netid.strip(), 'v4-fixed-ip': ''} for netid in kwargs['network_ids'].split(',') ]
            log.debug("Specifying networks for vm %s: %s",
                      node_name, str.join(', ', [nic['net-id'] for nic in nics]))

        vm = self.client.servers.create(
            node_name, image_id, flavor, key_name=key_name,
            security_groups=[security_group], userdata=image_userdata,
            nics=nics)

        self._instances[vm.id] = vm
        
        return vm.id

    def stop_instance(self, instance_id):
        """Stops the instance gracefully.

        :param str instance_id: instance identifier
        """
        instance = self._load_instance(instance_id)
        instance.delete()
        del self._instances[instance_id]

    def get_ips(self, instance_id):
        """Retrieves the private and public ip addresses for a given instance.

        :return: tuple (IPs)
        """
        self._load_instance(instance_id)
        instance = self._load_instance(instance_id)

        IPs = sum(instance.networks.values(), [])

        # We also need to check if there is any floating IP associated
        if self.request_floating_ip:
            # We need to list the floating IPs for this instance
            floating_ips = [ip for ip in self.client.floating_ips.list() if ip.instance_id == instance.id]
            if not floating_ips:
                log.debug("Public ip address has to be assigned through "
                          "elasticluster.")
                ip = self._allocate_address(instance)
                # This is probably the preferred IP we want to use
                IPs.insert(0, ip)
        return IPs

    def is_instance_running(self, instance_id):
        """Checks if the instance is up and running.

        :param str instance_id: instance identifier

        :return: bool - True if running, False otherwise
        """

        # Here, it's always better if we update the instance.        
        instance = self._load_instance(instance_id, force_reload=True)
        return instance.status == 'ACTIVE'

    # Protected methods

    def _check_keypair(self, name, public_key_path, private_key_path):
        """First checks if the keypair is valid, then checks if the keypair
        is registered with on the cloud. If not the keypair is added to the
        users ssh keys.

        :param str name: name of the ssh key
        :param str public_key_path: path to the ssh public key file
        :param str private_key_path: path to the ssh private key file

        :raises: `KeypairError` if key is not a valid RSA or DSA key,
                 the key could not be uploaded or the fingerprint does not
                 match to the one uploaded to the cloud.
        """

        # Read key. We do it as first thing because we need it either
        # way, to check the fingerprint of the remote keypair if it
        # exists already, or to create a new keypair.
        pkey = None
        try:
            pkey = DSSKey.from_private_key_file(private_key_path)
        except PasswordRequiredException:
            log.warning(
                "Unable to check key file `%s` because it is encrypted with a "
                "password. Please, ensure that you added it to the SSH agent "
                "with `ssh-add %s`", private_key_path, private_key_path)
        except SSHException:
            try:
                pkey = RSAKey.from_private_key_file(private_key_path)
            except PasswordRequiredException:
                log.warning(
                    "Unable to check key file `%s` because it is encrypted with a "
                    "password. Please, ensure that you added it to the SSH agent "
                    "with `ssh-add %s`", private_key_path, private_key_path)
            except SSHException:
                raise KeypairError('File `%s` is neither a valid DSA key '
                                   'or RSA key.' % private_key_path)

        try:
            # Check if a keypair `name` exists on the cloud.
            keypair = self.client.keypairs.get(name)

            # Check if it has the correct keypair, but only if we can read the local key
            if pkey:
                fingerprint = str.join(
                    ':', (i.encode('hex') for i in pkey.get_fingerprint()))
                if fingerprint != keypair.fingerprint:
                    raise KeypairError(
                        "Keypair `%s` is present but has "
                        "different fingerprint. Aborting!" % name)
            else:
                log.warning("Unable to check if the keypair is using the correct key.")
        except NotFound:
            log.warning(
                "Keypair `%s` not found on resource `%s`, Creating a new one",
                name, self._os_auth_url)

            # Create a new keypair
            with open(os.path.expanduser(public_key_path)) as f:
                key_material = f.read()
                try:
                    self.client.keypairs.create(name, key_material)
                except Exception, ex:
                    log.error(
                        "Could not import key `%s` with name `%s` to `%s`",
                        name, public_key_path, self._os_auth_url)
                    raise KeypairError(
                        "could not create keypair `%s`: %s" % (name, ex))


    def _check_security_group(self, name):
        """Checks if the security group exists.

        :param str name: name of the security group
        :raises: `SecurityGroupError` if group does not exist
        """
        security_groups = self.client.security_groups.list()

        # TODO: We should be able to create the security group if it
        # doesn't exist and at least add a rule to accept ssh access.
        # Also, we should be able to add new rules to a security group
        # if needed.
        if name not in [sg.name for sg in security_groups ]:
            raise SecurityGroupError(
                "the specified security group %s does not exist" % name)

    @memoize(120)
    def _get_images(self):
        """Get available images. We cache the results in order to reduce
        network usage.

        """
        return self.client.images.list()

    @memoize(120)
    def _get_flavors(self):
        """Get available flavors. We cache the results in order to reduce
        network usage.

        """
        return self.client.flavors.list()

    def _load_instance(self, instance_id, force_reload=True):
        """Checks if an instance with the given id is cached. If not it
        will connect to the cloud and put it into the local cache
        _instances.

        :param str instance_id: instance identifier
        :param bool force_reload: reload instance from server
        :return: py:class:`novaclient.v1_1.servers.Server` - instance
        :raises: `InstanceError` is returned if the instance can't
                 be found in the local cache or in the cloud.
        """
        if force_reload:
            try:
                # Remove from cache and get from server again
                vm = self.client.servers.get(instance_id)
                # update cache
                self._instances[instance_id] = vm
                # delete internal cache, just in case
                for i in self._cached_instances:
                    if i.id == instance_id:
                        self._cached_instances.remove(i)
                        self._cached_instances.append(vm)
                        break
                    
            except NotFound:
                raise InstanceError("the given instance `%s` was not found "
                                    "on the coud" % instance_id)
        if instance_id in self._instances:
            return self._instances[instance_id]

        # Instance not in the internal dictionary.
        # First, check the internal cache:
        if instance_id not in [i.id for i in self._cached_instances]:
            # Refresh the cache, just in case
            self._cached_instances = self.client.servers.list()

        for inst in self._cached_instances:
            if inst.id == instance_id:
                self._instances[instance_id] = inst
                return inst

        # If we reached this point, the instance was not found neither
        # in the cache or on the website.
        raise InstanceError("the given instance `%s` was not found "
                            "on the coud" % instance_id)

    def _allocate_address(self, instance):
        """Allocates a free public ip address to the given instance

        :param instance: instance to assign address to
        :type instance: py:class:`novaclient.v1_1.servers.Server`

        :return: public ip address
        """
        with OpenStackCloudProvider.__node_start_lock:
            free_ips = [i for i in self.client.floating_ips.list() if not i.fixed_ip]
            if not free_ips:
                free_ips.append(self.client.floating_ips.create())
            ip = free_ips.pop()
            instance.add_floating_ip(ip)
        return ip.ip

    # Fix pickler
    def __getstate__(self):
        return {'auth_url': self._os_auth_url,
                'username': self._os_username,
                'password': self._os_password,
                'project_name': self._os_tenant_name,
                'region_name': self._os_region_name,
                'request_floating_ip': self.request_floating_ip,
                'instance_ids': self._instances.keys(),
                'nova_api_version': self.nova_api_version,
            }

    def __setstate__(self, state):
        self._os_auth_url = state['auth_url']
        self._os_username = state['username']
        self._os_password = state['password']
        self._os_tenant_name = state['project_name']
        self._os_region_name = state['region_name']
        self.request_floating_ip = state['request_floating_ip']
        self.nova_api_version = state.get('nova_api_version', DEFAULT_OS_NOVA_API_VERSION)
        self.client = client.Client(self.nova_api_version,
                                    self._os_username, self._os_password, self._os_tenant_name,
                                    self._os_auth_url, region_name=self._os_region_name)
        self._instances = {}
        self._cached_instances = []
        # recover instance objects from the cloud.

        # This is Disabled as we don't want to reload the
        # instances also for "list" command
        
        # for vm in self.client.servers.list():
        #     if vm.id in state['instance_ids']:
        #         self._instances[vm.id] = vm

        # for missing_vm in set(state['instance_ids']).difference(self._instances.keys()):
        #     log.warning("VM with id %s not found in cloud "
        #                 "%s" % (missing_vm, self._os_auth_url))

        



#! /usr/bin/env python
#
# Copyright (C) 2013, 2014 GC3, University of Zurich
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = 'Nicolas Baer <nicolas.baer@uzh.ch>, Antonio Messina <antonio.s.messina@gmail.com>'

# System imports
import os
import re
import sys

# External modules
from ConfigParser import RawConfigParser

try:
    # Voluptuous version >= 0.8.1
    from voluptuous import message, MultipleInvalid, Invalid, Schema
    from voluptuous import All, Length, Any, Url, Boolean, Optional
except ImportError:
    # Voluptuous version <= 0.7.2
    from voluptuous.voluptuous import message, MultipleInvalid, Invalid
    from voluptuous import Schema, All, Length, Any, Url, Boolean, Optional

# Elasticluster imports
from elasticluster import log
from elasticluster.exceptions import ConfigurationError
from elasticluster.providers.ansible_provider import AnsibleSetupProvider
from elasticluster.cluster import Cluster
from elasticluster.repository import PickleRepository


class Configurator(object):
    """The `Configurator` class is responsible for

    1) keeping track of the configuration and

    2) offer factory methods to create all kind of objects that need
    information from the configuration.

    The cluster configuration dictionary is structured in the
    following way: (see an example @
    https://github.com/gc3-uzh-ch/elasticluster/wiki/Configuration-Module)::

          { "<cluster_template>" : {
              "setup" : { properties of the setup section },
              "cloud" : { properties of the cloud section },
              "login" : { properties of the login section },
              "cluster" : { properties of the cluster section },
              "nodes": {  "<node_kind>" : { properties of the node},
                          "<node_kind>" : { properties of the node},
                      },
              },
           "<cluster_template>" : {
              (see above)
              }
           }


    It is also responsible for loading a cluster from a valid
    `repository.AbstractClusterRepository`.

    :param dict cluster_conf: see description above

    :param str storage_path: path to store data

    :raises MultipleInvalid: configuration validation

    """

    setup_providers_map = {"ansible": AnsibleSetupProvider, }

    default_storage_dir = os.path.expanduser(
        "~/.elasticluster/storage")

    def __init__(self, cluster_conf, storage_path=None):
        self.general_conf = dict()
        self.cluster_conf = cluster_conf
        if storage_path:
            storage_path = os.path.expanduser(storage_path)
            storage_path = os.path.expandvars(storage_path)
            self.general_conf['storage'] = storage_path
        else:
            self.general_conf['storage'] = Configurator.default_storage_dir

        validator = ConfigValidator(self.cluster_conf)
        validator.validate()

    @classmethod
    def fromConfig(cls, configfiles, storage_path=None,
                   include_config_dirs=False):
        """Helper method to initialize Configurator from an ini file.

        :param str configfiles: path to the ini file(s). For each file
                                in `configfiles`, if a directory `*.d`
                                exists, also reads all the `*.conf`
                                files in that directory.

        :param str storage_path: path to the storage directory. If
                                 defined, a
                                 :py:class:`repository.PickleRepository`
                                 class will be instantiated.

        :param str include_config_dirs: Default is False. If True, for
                               each `file` in `configfiles` also files
                               in `file.d` ending with `.conf` will be
                               loaded

        :return: :py:class:`Configurator`

        """
        # FIXME: This is not python3 compatible!
        if isinstance(configfiles, basestring):
            configfiles = [configfiles]
        for cfgfile in configfiles[:]:
            cfgdir = cfgfile + '.d'
            if os.path.isfile(cfgfile) and os.path.isdir(cfgdir):
                for fname in os.listdir(cfgdir):
                    fullpath = os.path.join(cfgdir, fname)
                    if fname.endswith('.conf') and fullpath not in configfiles:
                        configfiles.append(fullpath)
        config_reader = ConfigReader(configfiles)
        conf = config_reader.read_config()
        return Configurator(conf, storage_path=storage_path)

    def create_cloud_provider(self, cluster_template):
        """Creates a cloud provider by inspecting the configuration properties
        of the given cluster template.

        :param str cluster_template: template to use (if not already specified
                                 on init)
        :return: cloud provider that fulfills the contract of
                 :py:class:`elasticluster.providers.AbstractSetupProvider`
        """
        conf = self.cluster_conf[cluster_template]['cloud']

        try:
            if conf['provider'] == 'ec2_boto':
                from elasticluster.providers.ec2_boto import BotoCloudProvider
                provider = BotoCloudProvider
            elif conf['provider'] == 'openstack':
                from elasticluster.providers.openstack import OpenStackCloudProvider
                provider = OpenStackCloudProvider
            elif conf['provider'] == 'google':
                from elasticluster.providers.gce import GoogleCloudProvider
                provider = GoogleCloudProvider
            elif conf['provider'] == 'azure':
                from elasticluster.providers.azure_provider import AzureCloudProvider
                provider = AzureCloudProvider
            else:
                raise Invalid("Invalid provider '%s' for cluster '%s'"% (conf['provider'], cluster_template))
        except ImportError, ex:
            raise Invalid("Unable to load provider '%s': %s" % (conf['provider'], ex))

        providerconf = conf.copy()
        providerconf.pop('provider')
        providerconf['storage_path'] = self.general_conf['storage']

        return provider(**providerconf)

    def create_cluster(self, template, name=None):
        """Creates a cluster by inspecting the configuration properties of the
        given cluster template.

        :param str template: name of the cluster template

        :param str name: name of the cluster. If not defined, the cluster
                         will be named after the template.

        :return: :py:class:`elasticluster.cluster.cluster` instance:

        :raises ConfigurationError: cluster template not found in config

        """
        if not name:
            name = template

        if template not in self.cluster_conf:
            raise ConfigurationError(
                "Invalid configuration for cluster `%s`: %s"
                "" % (template, name))

        conf = self.cluster_conf[template]
        conf_login = self.cluster_conf[template]['login']

        extra = conf['cluster'].copy()
        extra.pop('cloud')
        extra.pop('setup_provider')
        extra['template'] = template

        cluster = Cluster(name,
                          self.create_cloud_provider(template),
                          self.create_setup_provider(template, name=name),
                          conf_login['user_key_name'],
                          conf_login['user_key_public'],
                          conf_login["user_key_private"],
                          repository=self.create_repository(),
                          **extra)

        nodes = dict(
            (k[:-6], int(v)) for k, v in conf['cluster'].iteritems() if
            k.endswith('_nodes'))

        for kind, num in nodes.iteritems():
            conf_kind = conf['nodes'][kind]
            extra = conf_kind.copy()
            extra.pop('image_id', None)
            extra.pop('flavor', None)
            extra.pop('security_group', None)
            extra.pop('image_userdata', None)
            userdata = conf_kind.get('image_userdata', '')
            cluster.add_nodes(kind,
                              num,
                              conf_kind['image_id'],
                              conf_login['image_user'],
                              conf_kind['flavor'],
                              conf_kind['security_group'],
                              image_userdata=userdata,
                              **extra)
        return cluster

    def load_cluster(self, cluster_name):
        """Loads a cluster from the cluster repository.

        :param str cluster_name: name of the cluster
        :return: :py:class:`elasticluster.cluster.cluster` instance
        """
        repository = self.create_repository()
        cluster = repository.get(cluster_name)
        if not cluster._setup_provider:
            cluster._setup_provider = self.create_setup_provider(cluster.template)
        if not cluster._cloud_provider:
            cluster._cloud_provider = self.create_cloud_provider(cluster.template)
        cluster.update_config(
            self.cluster_conf[cluster.template]['cluster'],
            self.cluster_conf[cluster.template]['login']
        )
        return cluster

    def create_setup_provider(self, cluster_template, name=None):
        """Creates the setup provider for the given cluster template.

        :param str cluster_template: template of the cluster
        :param str name: name of the cluster to read configuration properties
        """
        conf = self.cluster_conf[cluster_template]['setup']
        conf['general_conf'] = self.general_conf.copy()
        if name:
            conf['cluster_name'] = name
        conf_login = self.cluster_conf[cluster_template]['login']

        provider_name = conf.get('provider')
        if provider_name not in Configurator.setup_providers_map:
            raise ConfigurationError(
                "Invalid value `%s` for `setup_provider` in configuration "
                "file." % provider_name)

        storage_path = self.general_conf['storage']
        if 'playbook_path' in conf:
            playbook_path = conf['playbook_path']
            del conf['playbook_path']
        else:
            playbook_path = None
        groups = dict((k[:-7], v.split(',')) for k, v
                      in conf.items() if k.endswith('_groups'))
        environment = dict()
        for nodekind, grps in groups.iteritems():
            if not isinstance(grps, list):
                groups[nodekind] = [grps]

            # Environment variables parsing
            environment[nodekind] = dict()
            for key, value in list(conf.items()) + list(self.cluster_conf[cluster_template]['cluster'].items()):
                # Set both group and global variables
                for prefix in ["%s_var_" % nodekind,
                               "global_var_"]:
                    if key.startswith(prefix):
                        var = key.replace(prefix, '')
                        environment[nodekind][var] = value
                        log.debug("setting variable %s=%s for node kind %s",
                                  var, value, nodekind)

        provider = Configurator.setup_providers_map[provider_name]
        return provider(groups, playbook_path=playbook_path,
                        environment_vars=environment,
                        storage_path=storage_path,
                        sudo_user=conf_login['image_user_sudo'],
                        sudo=conf_login['image_sudo'], **conf)

    def create_repository(self):
        storage_path = self.general_conf['storage']
        repository = PickleRepository(storage_path)
        return repository


class ConfigValidator(object):
    """Validator for the cluster configuration dictionary.

    :param config: dictionary containing cluster configuration properties
    """

    def __init__(self, config):
        self.config = config

    def _pre_validate(self):
        """Handles all pre validation phase functionality, such as:

        * reading environment variables
        * interpolating configuraiton options
        """
        # read cloud provider environment variables (ec2_boto, google, openstack, or azure)
        for cluster, props in self.config.iteritems():
            if "cloud" in props and "provider" in props['cloud']:

                for param, value in props['cloud'].iteritems():
                    PARAM = param.upper()
                    if not value and PARAM in os.environ:
                        props['cloud'][param] = os.environ[PARAM]

        # interpolate ansible path manual, since configobj does not offer
        # an easy way to handle this
        ansible_pb_dir = os.path.join(
            sys.prefix,
            'share/elasticluster/providers/ansible-playbooks')
        for cluster, props in self.config.iteritems():
            if 'setup' in props and 'playbook_path' in props['setup']:
                if props['setup']['playbook_path'].startswith(
                        "%(ansible_pb_dir)s"):
                    pbpath = props['setup']['playbook_path']
                    pbpath = pbpath.replace("%(ansible_pb_dir)s",
                                            str(ansible_pb_dir))
                    self.config[cluster]['setup']['playbook_path'] = pbpath

    def _post_validate(self):
        """Handles all post validation phase functionality, such as:

        * expanding file paths
        """
        # expand all paths
        for cluster, values in self.config.iteritems():
            conf = self.config[cluster]
            if 'playbook_path' in values['setup']:
                pbpath = os.path.expanduser(values['setup']['playbook_path'])
                conf['setup']['playbook_path'] = pbpath

            privkey = os.path.expanduser(values['login']['user_key_private'])
            conf['login']['user_key_private'] = privkey

            pubkey = os.path.expanduser(values['login']['user_key_public'])
            conf['login']['user_key_public'] = pubkey

    def validate(self):
        """Validates the given configuration :py:attr:`self.config` to comply
        with elasticluster. As well all types are converted to the expected
        format if possible.

        :raises: :py:class:`voluptuous.MultipleInvalid` if multiple
                 properties are not compliant
        :raises: :py:class:`voluptuous.Invalid` if one property is invalid
        """
        self._pre_validate()
        # custom validators
        @message("file could not be found")
        def check_file(v):
            f = os.path.expanduser(os.path.expanduser(v))
            if os.path.exists(f):
                return f
            else:
                raise Invalid("file could not be found `%s`" % v)

        @message("Unsupported nova API version")
        def nova_api_version(version):
            try:
                from novaclient import client,exceptions
                client.get_client_class(version)
                return version
            except exceptions.UnsupportedVersion as ex:
                raise Invalid(
                    "Invalid option for `nova_api_version`: %s" % ex)

        # schema to validate all cluster properties
        schema = {"cluster": {"cloud": All(str, Length(min=1)),
                              "setup_provider": All(str, Length(min=1)),
                              "login": All(str, Length(min=1)),
                          },
                  "setup": {"provider": All(str, Length(min=1)),
                            Optional("playbook_path"): check_file(),
                            Optional("ssh_pipelining"): Boolean(str),
                        },
                  "login": {"image_user": All(str, Length(min=1)),
                            "image_user_sudo": All(str, Length(min=1)),
                            "image_sudo": Boolean(str),
                            "user_key_name": All(str, Length(min=1)),
                            "user_key_private": check_file(),
                            "user_key_public": check_file(),
                        },
        }

        cloud_schema_ec2 = {"provider": 'ec2_boto',
                            "ec2_url": Url(str),
                            "ec2_access_key": All(str, Length(min=1)),
                            "ec2_secret_key": All(str, Length(min=1)),
                            "ec2_region": All(str, Length(min=1)),
                            Optional("request_floating_ip"): Boolean(str),
                            Optional("vpc"): All(str, Length(min=1)),
        }
        cloud_schema_gce = {"provider": 'google',
                            "gce_client_id": All(str, Length(min=1)),
                            "gce_client_secret": All(str, Length(min=1)),
                            "gce_project_id": All(str, Length(min=1)),
                            Optional("noauth_local_webserver"): Boolean(str),
                            Optional("zone"): All(str, Length(min=1)),
        }

        cloud_schema_openstack = {"provider": 'openstack',
                                  "auth_url": All(str, Length(min=1)),
                                  "username": All(str, Length(min=1)),
                                  "password": All(str, Length(min=1)),
                                  "project_name": All(str, Length(min=1)),
                                  Optional("request_floating_ip"): Boolean(str),
                                  Optional("region_name"): All(str, Length(min=1)),
                                  Optional("nova_api_version"): nova_api_version(),
        }
        cloud_schema_azure = {"provider": 'azure',
                              "subscription_id": All(str, Length(min=1)),
                              "certificate": All(str, Length(min=1)),
        }
        node_schema = {
            "flavor": All(str, Length(min=1)),
            "image_id": All(str, Length(min=1)),
            Optional("security_group"): All(str, Length(min=1)),
            Optional("network_ids"): All(str, Length(min=1)),
        }

        # validation
        validator = Schema(schema, required=True, extra=True)
        node_validator = Schema(node_schema, required=True, extra=True)
        ec2_validator = Schema(cloud_schema_ec2, required=True, extra=False)
        gce_validator = Schema(cloud_schema_gce, required=True, extra=False)
        openstack_validator = Schema(cloud_schema_openstack, required=True, extra=False)
        azure_validator = Schema(cloud_schema_azure, required=True, extra=False)

        if not self.config:
            raise Invalid("No clusters found in configuration.")

        for cluster, properties in self.config.iteritems():
            self.config[cluster] = validator(properties)

            if 'provider' not in properties['cloud']:
                raise Invalid(
                    "Missing `provider` option in cluster `%s`" % cluster)
            try:
                cloud_props = properties['cloud']
                if properties['cloud']['provider'] == "ec2_boto":
                    self.config[cluster]['cloud'] = ec2_validator(cloud_props)
                elif properties['cloud']['provider'] == "google":
                    self.config[cluster]['cloud'] = gce_validator(cloud_props)
                elif properties['cloud']['provider'] == "openstack":
                    self.config[cluster]['cloud'] = openstack_validator(cloud_props)
                elif properties['cloud']['provider'] == "azure":
                    self.config[cluster]['cloud'] = azure_validator(cloud_props)
            except MultipleInvalid as ex:
                raise Invalid("Invalid configuration for cloud section `cloud/%s`: %s" % (properties['cluster']['cloud'], str.join(", ", [str(i) for i in ex.errors])))


            if 'nodes' not in properties or len(properties['nodes']) == 0:
                raise Invalid(
                    "No nodes configured for cluster `%s`" % cluster)

            for node, props in properties['nodes'].iteritems():
                # check name pattern to conform hostnames
                match = re.search(r'^[a-zA-Z0-9-]*$', node)
                if not match:
                    raise Invalid(
                        "Invalid name `%s` for node group. A valid node group"
                        " can only consist of letters, digits or the hyphen"
                        " character (`-`)" % (node,))

                node_validator(props)

                if (properties['cloud']['provider'] == 'ec2_boto'
                    and 'vpc' in self.config[cluster]['cloud']
                    and 'network_ids' not in props):
                    raise Invalid(
                        "Node group `%s/%s` is being used in"
                        " a VPC, so it must specify network_ids."
                        % (cluster, node))

                if (properties['cloud']['provider'] == 'ec2_boto'
                    and 'network_ids' in props
                    and 'vpc' not in self.config[cluster]['cloud']):
                    raise Invalid(
                        "Cluster `%s` must specify a VPC to place"
                        " `%s` instances in %s"
                        % (cluster, node, props['network_ids']))

        self._post_validate()


class ConfigReader(object):
    """Reads the configuration properties from a ini file.

    :param str configfile: path to configfile
    """
    cluster_section = "cluster"
    login_section = "login"
    setup_section = "setup"
    cloud_section = "cloud"
    node_section = "node"

    def __init__(self, configfiles):
        self.configfiles = configfiles

        configparser = RawConfigParser()
        config_tmp = configparser.read(self.configfiles)
        self.conf = dict()
        for section in configparser.sections():
            self.conf[section] = dict(configparser.items(section))

        #self.conf = ConfigObj(self.configfile, interpolation=False)

        @message("file could not be found")
        def check_file(v):
            f = os.path.expanduser(os.path.expanduser(v))
            if os.path.exists(f):
                return f
            else:
                raise Invalid("file could not be found `%s`" % v)

        @message("Unsupported nova API version")
        def nova_api_version(version):
            try:
                from novaclient import client, exceptions
                client.get_client_class(version)
                return version
            except exceptions.UnsupportedVersion as ex:
                raise Invalid(
                    "Invalid option for `nova_api_version`: %s" % ex)

        self.schemas = {
            "cloud": Schema(
                {"provider": Any('ec2_boto', 'google', 'openstack', 'azure'),
                 "ec2_url": Url(str),
                 "ec2_access_key": All(str, Length(min=1)),
                 "ec2_secret_key": All(str, Length(min=1)),
                 "ec2_region": All(str, Length(min=1)),
                 "auth_url": All(str, Length(min=1)),
                 "username": All(str, Length(min=1)),
                 "password": All(str, Length(min=1)),
                 "tenant_name": All(str, Length(min=1)),
                 Optional("region_name"): All(str, Length(min=1)),
                 "gce_project_id": All(str, Length(min=1)),
                 "gce_client_id": All(str, Length(min=1)),
                 "gce_client_secret": All(str, Length(min=1)),
                 "nova_client_api": nova_api_version()}, extra=True),
                 "subscription_id": All(str, Length(min=1)),
                 "certificate": All(str, Length(min=1)),
                 "cloud_service_name": All(str, Length(min=1)),
                 "location": All(str, Length(min=1)),
            "cluster": Schema(
                {"cloud": All(str, Length(min=1)),
                 "setup_provider": All(str, Length(min=1)),
                 "login": All(str, Length(min=1))}, required=True, extra=True),
                  "cloud_service_name": All(str, Length(min=1)),
                  "location": All(str, Length(min=1)),
            "setup": Schema(
                {"provider": All(str, Length(min=1)),
                    }, required=True, extra=True),
            "login": Schema(
                {"image_user": All(str, Length(min=1)),
                 "image_user_sudo": All(str, Length(min=1)),
                 "image_sudo": Boolean(str),
                 "user_key_name": All(str, Length(min=1)),
                 "user_key_private": check_file(),
                 "user_key_public": check_file()}, required=True)
        }

    def read_config(self):
        """Reads the configuration properties from the ini file and links the
        section to comply with the cluster config dictionary format.

        :return: dictionary containing all configuration properties from the
         ini file in compliance to the cluster config format
        :raises: :py:class:`voluptuous.MultipleInvalid` if not all sections
                 present or broken links between secitons
        """
        clusters = dict((key, value) for key, value in self.conf.iteritems() if
                        re.search(ConfigReader.cluster_section + "/(.*)", key)
                        and key.count("/") == 1)

        conf_values = dict()

        errors = MultipleInvalid()
        # FIXME: to be refactored:
        # we should check independently each one of the sections, and raise errors accordingly.

        for cluster in clusters:
            # Get the name of the cluster
            name = re.search(ConfigReader.cluster_section + "/(.*)",
                             cluster).groups()[0]
            if not name:
                errors.add("Invalid section name `%s`" % cluster)
                continue

            cluster_conf = dict(self.conf[cluster])

            try:
                self.schemas['cluster'](cluster_conf)
            except MultipleInvalid as ex:
                for error in ex.errors:
                    errors.add("Section `%s`: %s" % (cluster, error))
                continue

            cloud_name = ConfigReader.cloud_section + "/" + cluster_conf[
                'cloud']
            login_name = ConfigReader.login_section + "/" + cluster_conf[
                'login']
            setup_name = ConfigReader.setup_section + "/" + cluster_conf[
                'setup_provider']

            values = dict()
            values['cluster'] = cluster_conf
            try:
                values['setup'] = dict(self.conf[setup_name])
                self.schemas['setup'](values['setup'])
            except KeyError, ex:
                errors.add(
                    "cluster `%s` setup section `%s` does not exists" % (
                        cluster, setup_name))
            except MultipleInvalid, ex:
                for error in ex.errors:
                    errors.add(error)

            try:
                values['login'] = dict(self.conf[login_name])
                self.schemas['login'](values['login'])
            except KeyError, ex:
                errors.add(
                    "cluster `%s` login section `%s` does not exists" % (
                        cluster, login_name))
            except MultipleInvalid, ex:
                errors.add(Invalid("Error in login section `%s`: %s" % (
                    login_name, str.join(', ', [str(e) for e in ex.errors]))))

            try:
                values['cloud'] = dict(self.conf[cloud_name])
                self.schemas['cloud'](values['cloud'])
            except KeyError, ex:
                errors.add(
                    "cluster `%s` cloud section `%s` does not exists" % (
                        cluster, cloud_name))
            except MultipleInvalid, ex:
                for error in ex.errors:
                    errors.add(Invalid("section %s: %s" % (cloud_name, error)))

            try:
                # nodes can inherit the properties of cluster or overwrite them
                nodes = dict((key, value) for key, value in
                             values['cluster'].items() if
                             key.endswith('_nodes'))
                values['nodes'] = dict()
                for node in nodes.iterkeys():
                    node_name = re.search("(.*)_nodes", node).groups()[0]
                    property_name = "%s/%s/%s" % (ConfigReader.cluster_section,
                                                  name, node_name)
                    if property_name in self.conf:
                        node_values = dict(
                            (key, value.strip("'").strip('"')) for key, value
                            in self.conf[property_name].iteritems())
                        node_values = dict(
                            values['cluster'].items() + node_values.items())
                        values['nodes'][node_name] = node_values
                    else:
                        values['nodes'][node_name] = values['cluster']

                if errors.errors:
                    log.error("Ignoring cluster `%s`: %s" % (
                        name, str.join(", ", [str(e) for e in errors.errors])))
                else:
                    conf_values[name] = values
            except KeyError, ex:
                errors.add("Error in section `%s`" % cluster)

        # FIXME: do we really need to raise an exception if we cannot
        # parse *part* of the configuration files? We should just
        # ignore those with errors and return both the parsed
        # configuration values _and_ a list of errors
        if errors.errors:
            raise errors
        return conf_values



#
# Copyright (C) 2013 GC3, University of Zurich
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__author__ = 'Nicolas Baer <nicolas.baer@uzh.ch>'


# stdlib imports
from abc import ABCMeta, abstractmethod


class AbstractCloudProvider:
    """Defines the contract for a cloud provider to proper function with
    elasticluster.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, **config):
        """The constructor of a `CloudProvider` class is called only
        using keyword arguments.

        Usually these are configuration option of the corresponding
        `setup` section in the configuration file.
        """
        pass

    @abstractmethod
    def start_instance(self, key_name, public_key_path, private_key_path,
                       security_group, flavor, image_id, image_userdata,
                       username=None, node_name=None):
        """Starts a new instance on the cloud using the given properties.
        Multiple instances might be started in different threads at the same
        time. The implementation should handle any problems regarding this
        itself.

        :param str key_name: name of the ssh key to connect
        :param str public_key_path: path to ssh public key
        :param str private_key_path: path to ssh private key
        :param str security_group: firewall rule definition to apply on the
                                   instance
        :param str flavor: machine type to use for the instance
        :param str image_name: image type (os) to use for the instance
        :param str image_userdata: command to execute after startup
        :param str username: username for the given ssh key, default None

        :return: str - instance id of the started instance
        """
        pass

    @abstractmethod
    def stop_instance(self, instance_id):
        """Stops the instance gracefully.

        :param str instance_id: instance identifier

        :return: None
        """
        pass

    @abstractmethod
    def get_ips(self, instance_id):
        """Retrieves the private and public ip addresses for a given instance.

        :return: list (IPs)
        """
        pass

    @abstractmethod
    def is_instance_running(self, instance_id):
        """Checks if the instance is up and running.

        :param str instance_id: instance identifier

        :return: bool - True if running, False otherwise
        """
        pass


class AbstractSetupProvider:
    """
    TODO: define...
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def setup_cluster(self, cluster):
        """Configures all nodes of a cluster to function in respect to the
        given configuration.

        This method *must* be idempotent, i.e. it should always be
        safe calling it multiple times..

        :param cluster: cluster to configure
        :type cluster: :py:class:`elasticluster.cluster.Cluster`

        :return: `True` if the cluster is correctly configured, even
                  if the method didn't actually do anything. `False` if the
                  cluster is not configured.
        """
        pass

    @abstractmethod
    def cleanup(self):
        """Cleanup any temporary file or directory created during setup.
        This method is called every time a cluster is stopped.

        :return: None
        """
        pass



#!/usr/bin/env python
# Azure provider for elasticluster

# elasticluster 'azure' package conflicts with azure SDK. This fixes
# it by causing "import azure" to look for a system library module.
from __future__ import absolute_import

# System imports
from math import floor, ceil
import base64
import subprocess
import time
import re
import threading
import xml.etree.ElementTree as xmltree

# External imports
import azure
import azure.servicemanagement
from azure.http import HTTPRequest

# Elasticluster imports
from elasticluster import log
from elasticluster.providers import AbstractCloudProvider
from elasticluster.exceptions import CloudProviderError

SSH_PORT = 22
PORT_MAP_OFFSET = 1200
DEFAULT_WAIT_TIMEOUT = 600
WAIT_RESULT_SLEEP = 10
VNET_NS = 'http://schemas.microsoft.com/ServiceHosting' \
          '/2011/07/NetworkConfiguration'

# resource-management constants
VMS_PER_CLOUD_SERVICE = 20
VMS_PER_STORAGE_ACCOUNT = 40
VMS_PER_VNET = 2000
CLOUD_SERVICES_PER_SUBSCRIPTION = 20

# helper functions


def _run_command(args):
    p = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    return p.returncode, stdout, stderr


def _check_positive_integer(name, value):
    ret = -1
    try:
        if '.' in value:
            raise ValueError
        ret = int(value)
        if ret < 1:
            raise ValueError
    except Exception:
        err = "invalid value '%s' for %s, must be integer > 0" % (
            value, name)
        log.error(err)
        raise Exception(err)
    return ret


def ceil_div(a, b):
    return int(ceil(float(a) / b))


def floor_div(a, b):
    return int(floor(float(a) / b))


def _rest_put(subscription, path, xml):
    # can't use SDK _perform_put because we need text/plain content-type
    request = HTTPRequest()
    request.method = 'PUT'
    request.host = azure.MANAGEMENT_HOST
    request.path = path
    request.body = azure._get_request_body(xml)
    request.path, request.query = azure._update_request_uri_query(request)
    # request.headers.append(('Content-Length', str(len(request.body))))
    request.headers.append(('Content-Type', 'text/plain'))
    request.headers = subscription._sms._update_management_header(
        request, azure.servicemanagement.X_MS_VERSION)
    response = subscription._sms._perform_request(request)
    return response


class AzureGlobalConfig:

    """Manage all the settings for an Azure cluster which are
    global as opposed to node-specific.

    This does not manage resources (which are dynamic), only settings
    (which are static, or at least are determined at cluster startup).
    """

    # we only get a few pieces of information at this point
    def __init__(self, parent, subscription_id, certificate, storage_path):
        self._setup_done = False

        self._parent = parent
        # if we got a subscription here, use it. if not,
        # a subscription_file has to be
        # included in the cluster parameters passed to setup().
        if subscription_id:
            self._parent._subscriptions.append(
                AzureSubscription(
                    config=self,
                    subscription_id=subscription_id,
                    certificate=certificate, index=0))

        self._key_name = None
        self._public_key_path = None
        self._private_key_path = None
        self._location = None
        self._base_name = None
        self._username = None
        self._subscription_file = None
        self._use_public_ips = None
        self._n_cloud_services = None
        self._n_storage_accounts = None
        self._n_vms_requested = None
        self._n_subscriptions = None

    # called when the first node is about to be created. this is where
    # we get most of the config info
    def setup(
            self,
            key_name,
            public_key_path,
            private_key_path,
            security_group,
            location,
            base_name=None,
            username=None,
            subscription_file=None,
            frontend_nodes=None,
            compute_nodes=None,
            use_public_ips=None,
            wait_timeout=DEFAULT_WAIT_TIMEOUT,
            n_cloud_services=None,
            n_storage_accounts=None,
            **kwargs):
        self._setup_done = True
        self._key_name = key_name
        self._public_key_path = public_key_path
        self._private_key_path = private_key_path
        self._security_group = security_group
        self._location = location
        self._base_name = base_name
        self._username = username
        self._subscription_file = subscription_file
        self._wait_timeout = _check_positive_integer(
            'wait_timeout', wait_timeout)

        # elasticluster parser doesn't know about bools
        self._use_public_ips = True if use_public_ips == 'True' else False

        # subscriptions
        self._read_subscriptions()

        # resource names will be generated based on this string.
        if len(self._base_name) < 3 or len(self._base_name) > 15:
            err = 'base_name %s not between 3 and 15 characters' \
                  % self._base_name
            log.error(err)
            raise Exception(err)
        if re.match('[^a-z0-9]', self._base_name):
            err = 'base_name %s is invalid, only lowercase letters and ' \
                  'digits allowed' % self._base_name
            log.error(err)
            raise Exception(err)

        n_frontend_nodes = n_compute_nodes = 0
        if frontend_nodes:
            n_frontend_nodes = _check_positive_integer('frontend_nodes',
                                                       frontend_nodes)
        if compute_nodes:
            n_compute_nodes = _check_positive_integer('compute_nodes',
                                                      compute_nodes)
        self._n_vms_requested = n_frontend_nodes + n_compute_nodes

        # if we got resource counts, read them. otherwise compute them.
        if n_cloud_services:
            self._n_cloud_services = _check_positive_integer(
                'n_cloud_services', n_cloud_services)
        else:
            self._n_cloud_services = ceil_div(self._n_vms_requested,
                                              VMS_PER_CLOUD_SERVICE)

        if n_storage_accounts:
            self._n_storage_accounts = _check_positive_integer(
                'n_storage_accounts', n_storage_accounts)
        else:
            self._n_storage_accounts = ceil_div(self._n_vms_requested,
                                                VMS_PER_STORAGE_ACCOUNT)

        self._n_subscriptions = len(self._parent._subscriptions)
        min_subscriptions = ceil_div(self._n_cloud_services,
                                     CLOUD_SERVICES_PER_SUBSCRIPTION)
        if self._n_subscriptions < min_subscriptions:
            err = 'Not enough subscriptions available to meet resource' \
                  'requirements (have %i, need %i)' % \
                  (len(self._parent._subscriptions), min_subscriptions)
            log.error(err)
            raise Exception(err)
        log.debug('compute nodes: %i. cloud services: %i. storage accounts:'
                  ' %i. subscriptions: %i.', self._n_vms_requested,
                  self._n_cloud_services, self._n_storage_accounts,
                  self._n_subscriptions)

        # store values needed to map vms to resources
        self._vm_per_cs = ceil_div(
            self._n_vms_requested, self._n_cloud_services)
        self._cs_per_sub = ceil_div(
            self._n_cloud_services, self._n_subscriptions)
        self._vm_per_sub = self._vm_per_cs * self._cs_per_sub

        self._vm_per_sa = ceil_div(
            self._n_vms_requested, self._n_storage_accounts)
        self._sa_per_sub = ceil_div(
            self._n_storage_accounts, self._n_subscriptions)
        if self._vm_per_sub != self._vm_per_sa * self._sa_per_sub:
            err = 'inconsistency, %i != %i' % \
                  (self._vm_per_sub, self._vm_per_sa * self._sa_per_sub)
            log.error(err)
            raise Exception(err)

    # there are three ways to designate a node in the cluster:
    # - by its flat index (0..total nodes - 1)
    # - by its node name (we will use basename_vm%i where %i is the flat index)
    # - by its resources - index of its subscription, its cloud service
    #   (or alternatively its storage account), and its index within that
    #   cloud service or storage account.
    # We need all possible conversions amongst these.

    def _vm_name_to_flat(self, vm_name):
        if vm_name and vm_name.startswith(self._base_name + '_vm'):
            bare = vm_name[len(self._base_name) + 3:]
            return int(bare)
        err = '_vm_name_to_flat: can't process vm_name %s' % vm_name
        log.error(err)
        raise Exception(err)

    def _vm_flat_to_resources(self, vm_index):
        i_sub = floor_div(vm_index, self._vm_per_sub)
        i_cs = floor_div(vm_index % self._vm_per_sub, self._vm_per_cs)
        i_vm_in_cs = vm_index - (i_sub * self._vm_per_sub) - \
            (i_cs * self._vm_per_cs)
        i_sa = floor_div(vm_index % self._vm_per_sub, self._vm_per_sa)
        i_vm_in_sa = vm_index - (i_sub * self._vm_per_sub) - \
            (i_sa * self._vm_per_sa)
        return i_sub, i_cs, i_vm_in_cs, i_sa, i_vm_in_sa

    def _vm_name_to_resources(self, vm_name):
        return self._vm_flat_to_resources(self._vm_name_to_flat(vm_name))

    def _read_subscriptions(self):
        ids = set()
        if self._subscription_file:
            if self._parent._subscriptions:
                log.debug(
                    'subscription was passed in cloud section, and '
                    'subscription_file was '
                    'also passed in cluster section. using both.')
                ids.add(self._parent._subscriptions[0]._subscription_id)
            try:
                pattern = re.compile(
                    'subscription_id=(\S*)\s+certificate=(.*)')
                with open(self._subscription_file) as f:
                    for line in f:
                        m = pattern.match(line)
                        if m:
                            if m.group(1) in ids:
                                log.debug('subscription_id %s has already been'
                                          ' read. ignoring.' % m.group(1))
                            else:
                                ids.add(m.group(1))
                                index = len(self._parent._subscriptions)
                                self._parent._subscriptions.append(
                                    AzureSubscription(config=self,
                                                      subscription_id=m.group(
                                                          1),
                                                      certificate=m.group(2),
                                                      index=index))
                        else:
                            log.debug(
                                'ignoring line in subscription_file %s: %s' %
                                (self._subscription_file, line))
            except Exception as e:
                log.error('error parsing subscription file %s: %s' %
                          (self._subscription_file, e))


class AzureSubscription:

    def __init__(self, config, subscription_id, certificate, index):
        self._do_cleanup = False    # never delete a subscription
        self._config = config
        self._subscription_id = subscription_id
        self._certificate = certificate
        self._index = index
        self._sms_internal = None
        self._resource_lock_internal = None
        self._storage_accounts = list()
        self._cloud_services = list()
        self._vnet = None

        rc, stdout, stderr = _run_command(
            ['openssl', 'x509', '-in', self._certificate,
             '-fingerprint', '-noout'])
        if rc != 0:
            err = "error getting fingerprint: %s" % stderr
            log.error(err)
            raise Exception(err)
        self._fingerprint = stdout.strip()[17:].replace(':', '')

        rc, stdout, stderr = _run_command(
            ['openssl', 'pkcs12', '-export', '-in', self._certificate,
             '-nokeys', '-password', 'pass:'])
        if rc != 0:
            err = "error getting pkcs12 signature: %s" % stderr
            log.error(err)
            raise Exception(err)
        self._pkcs12_base64 = base64.b64encode(stdout.strip())

    @property
    def _resource_lock(self):
        if self._resource_lock_internal is None:
            self._resource_lock_internal = threading.Lock()
        return self._resource_lock_internal

    @property
    def _sms(self):
        with self._resource_lock:
            if self._sms_internal is None:
                try:
                    self._sms_internal = \
                        azure.servicemanagement.ServiceManagementService(
                            self._subscription_id, self._certificate)
                except Exception as e:
                    log.error('error initializing azure serice: %s' % e)
                    raise
            return self._sms_internal

    @property
    def _n_instances(self):
        ret = 0
        for cloud_service in self._cloud_services:
            ret = ret + cloud_service._n_instances
        return ret

    # this needs to be in here because sms is per-subscription
    def _wait_result(self, req, timeout=None):
        if not timeout:
            timeout = self._config._wait_timeout
        if not req:
            return  # sometimes this happens, seems to mean success
        giveup_time = time.time() + timeout
        while giveup_time > time.time():
            operation_result = self._sms.get_operation_status(req.request_id)
            if operation_result.status == "InProgress":
                time.sleep(WAIT_RESULT_SLEEP)
                continue
            if operation_result.status == "Succeeded":
                return
            if operation_result.status == "Failed":
                err = 'async operation failed: %s' \
                      % operation_result.error.message
                log.error(err)
                raise CloudProviderError(err)
        err = 'async operation timed out'
        log.error(err)
        raise CloudProviderError(err)

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_resource_lock_internal']
        del d['_sms_internal']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self._resource_lock_internal = None
        self._sms_internal = None

    def _find_os_disks(self):
        # associate disk names of os vhd's with their nodes.
        try:
            disks = self._sms.list_disks()
            for disk in disks:
                for cloud_service in self._cloud_services:
                    for vm_name, vm in cloud_service._instances.iteritems():
                        if vm._os_vhd_name:
                            continue
                        if vm_name in disk.name and \
                                self._config._base_name in disk.name:
                            vm._os_vhd_name = disk.name
        except Exception as e:
            log.error('error in _find_os_disks: %s' % e)
            raise


class AzureCloudService:

    def __init__(self, config, subscription, index):
        self._do_cleanup = False
        self._config = config
        self._subscription = subscription
        self._location = config._location
        self._index = index
        self._name = "%s0su%ics%i" % (
            self._config._base_name, self._subscription._index, self._index)
        # treat deployment as sub-item of cloud service
        # deployment has same name as cloud service
        # only instances owned by this cloud service:
        self._instances = {}
        self._resource_lock_internal = None

    @property
    def _resource_lock(self):
        if self._resource_lock_internal is None:
            self._resource_lock_internal = threading.Lock()
        return self._resource_lock_internal

    @property
    def _n_instances(self):
        with self._resource_lock:
            return len(self._instances)

    @property
    def _exists(self):
        try:
            result = self._subscription._sms.get_hosted_service_properties(
                service_name=self._name)
            return True    # already exists
        except Exception as e:
            if str(e) != 'Not found (Not Found)':
                log.error('error checking for cloud service %s: %s' %
                          (self._name, str(e)))
                raise
        return False

    def _create(self):
        if not self._exists:
            try:
                result = self._subscription._sms.create_hosted_service(
                    service_name=self._name,
                    label=self._name,
                    location=self._location)
                self._subscription._wait_result(result)
            except Exception as e:
                # this shouldn't happen
                # if str(e) == 'Conflict (Conflict)':
                #    return False
                log.error('error creating cloud service %s: %s' %
                          (self._name, e))
                raise

    def _delete(self):
        if self._deployment:
            log.error("can't delete cloud service %s. It contains a "
                      "deployment and at least one node." % self._name)
            return False
        try:
            self._subscription._sms.delete_hosted_service(
                service_name=self._name)
        except Exception as e:
            log.error('error deleting cloud service %s: %s' %
                      (self._name, e))
            raise
        return True

    # don't call directly. called by an AzureVM when it's being deleted
    # and is the last node in the deployment.
    def _delete_deployment(self):
        try:
            result = self._subscription._sms.delete_deployment(
                service_name=self._name,
                deployment_name=self._name)
            self._subscription._wait_result(result)
        except Exception as e:
            log.error('error deleting deployment from cloud service %s: %s' %
                      (self._name, e))
            raise

    @property
    def _deployment(self):
        try:
            dep = self._subscription._sms.get_deployment_by_name(
                service_name=self._name, deployment_name=self._name)
            return dep
        except Exception as exc:
            if str(exc) == 'Not found (Not Found)':
                return None
            log.error('error getting deployment %s: %s' % (self._name, exc))
            raise

    def _add_certificate(self):
        # Add certificate to cloud service
        result = self._subscription._sms.add_service_certificate(
            self._name, self._subscription._pkcs12_base64, 'pfx', '')
        self._subscription._wait_result(result)

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_resource_lock_internal']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self._resource_lock_internal = None


class AzureStorageAccount:

    def __init__(self, config, subscription, index):
        self._do_cleanup = False
        self._created = False
        self._config = config
        self._subscription = subscription
        self._index = index
        self._name = "%s0su%ist%i" % (
            self._config._base_name, self._subscription._index, self._index)
        self._resource_lock_internal = None

    @property
    def _resource_lock(self):
        if self._resource_lock_internal is None:
            self._resource_lock_internal = threading.Lock()
        return self._resource_lock_internal

    def _exists(self):
        try:
            result = self._subscription._sms.get_storage_account_properties(
                service_name=self._name)
            # TODO dsteinkraus - make sure it's actually one of ours
            return True
        except Exception as e:
            if str(e) != 'Not found (Not Found)':
                log.error('error checking for storage account %s: %s' %
                          (self._name, str(e)))
                raise
            return False

    def _create(self):
        if not self._exists():
            try:
                result = self._subscription._sms.create_storage_account(
                    service_name=self._name,
                    description=self._name,
                    label=self._name,
                    location=self._config._location,
                    account_type='Standard_LRS'
                )
                # this seems to be taking much longer than the others...
                self._subscription._wait_result(
                    result, self._config._wait_timeout * 10)
            except Exception as e:
                # this shouldn't happen
                # if str(e) == 'Conflict (Conflict)':
                #    return False
                log.error('error creating storage account: %s' % str(e))
                raise

    def _delete(self):
        try:
            self._subscription._sms.delete_storage_account(
                service_name=self._name)
        except Exception as exc:
            log.error('error deleting storage account %s: %s' %
                      (self._name, exc))
            raise

    def _create_vhd(self, node_name, image_id):
        disk_url = u'http://%s.blob.core.windows.net/vhds/%s.vhd' % (
            self._name, node_name)
        vhd = azure.servicemanagement.OSVirtualHardDisk(image_id, disk_url)
        return vhd, disk_url

    def _delete_vhd(self, name):
        attempts = 100
        for attempt in range(1, attempts):
            try:
                # delete_vhd=False doesn't seem to help if the disk is not
                # ready to be deleted yet
                self._subscription._sms.delete_disk(
                    disk_name=name, delete_vhd=True)
                log.debug('_delete_vhd %s: success on attempt %i' %
                          (name, attempt))
                return
            except Exception as e:
                if str(e) == 'Not found (Not Found)':
                    log.debug(
                        "_delete_vhd: 'not found' deleting %s, assuming "
                        "success" % name)
                    return
                # log.error('_delete_vhd: error on attempt #%i to delete disk
                # %s: %s' % (attempt, name, e))
                time.sleep(10)
        err = '_delete_vhd %s: giving up after %i attempts' % (name, attempts)
        log.error(err)
        raise Exception(err)

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_resource_lock_internal']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self._resource_lock_internal = None


class AzureVNet:

    def __init__(self, config, subscription, index):
        self._do_cleanup = False
        self._config = config
        self._subscription = subscription
        self._index = index
        self._name = "%s0su%ivn%i" % (
            self._config._base_name, self._subscription._index, self._index)
        self._resource_lock_internal = None

    @property
    def _resource_lock(self):
        if self._resource_lock_internal is None:
            self._resource_lock_internal = threading.Lock()
        return self._resource_lock_internal

    def _exists(self):
        try:
            result = self._subscription._sms.list_virtual_network_sites()
            if len(result):
                for virtual_network_site in result.virtual_network_sites:
                    if virtual_network_site.name == self._name:
                        return True
                return False
            else:
                return False    # no vnets
        except Exception as exc:
            log.error('error checking existence of vnet %s: %s',
                      self._name, exc)
            raise

    def _create(self):
        if not self._exists():
            try:
                # note that this is replacing the whole network config,
                # so would succeed even if vnet already existed
                path = "/%s/services/networking/media" % \
                       self._subscription._subscription_id
                xml = self._create_vnet_to_xml(
                    location=self._config._location, vnet_name=self._name)
                result = _rest_put(self._subscription, path, xml)
                log.debug('created vnet %s', self._name)
            except Exception as e:
                err = 'error in _create_virtual_network: %s' % e
                log.error(err)
                raise

    def _delete(self):
        if self._exists():
            # TODO dsteinkraus - to have full support for creating and
            # deleting a vnet while a subscription has other vnets, we
            # would need to download current config, remove or add the
            # vnet we want to change, and re-upload the config. Not sure
            # how useful this would be, since it's easy to create a new
            # subscription for a project and just use one vnet in that
            # subscription.
            return

    def _deco(self, str):
        return '{%s}%s' % (VNET_NS, str)

    def _create_vnet_to_xml(self, location, vnet_name):
        try:
            template = "<NetworkConfiguration xmlns:xsd=\"http://www.w3.org/" \
                       "2001/XMLSchema\" xmlns:xsi=\"http://www.w3.org/2001/" \
                       "XMLSchema-instance\" xmlns=\"http://schemas." \
                       "microsoft.com/ServiceHosting/2011/07/" \
                       "NetworkConfiguration\">"
            template = template + """
  <VirtualNetworkConfiguration>
    <Dns>
      <DnsServers>
      </DnsServers>
    </Dns>
    <VirtualNetworkSites>
      <VirtualNetworkSite name="" Location="">
        <AddressSpace>
          <AddressPrefix>10.0.0.0/8</AddressPrefix>
        </AddressSpace>
        <Subnets>
          <Subnet name="subnet1">
            <AddressPrefix>10.0.0.0/11</AddressPrefix>
          </Subnet>
        </Subnets>
        <DnsServersRef>
        </DnsServersRef>
      </VirtualNetworkSite>
    </VirtualNetworkSites>
  </VirtualNetworkConfiguration>
</NetworkConfiguration>"""
            xmltree.register_namespace('', 'http://www.w3.org/2001/XMLSchema')
            xmltree.register_namespace(
                '', 'http://www.w3.org/2001/XMLSchema-instance')
            xmltree.register_namespace('', VNET_NS)
            tree = xmltree.fromstring(template)
            config = tree.find(self._deco('VirtualNetworkConfiguration'))
            sites = config.find(self._deco('VirtualNetworkSites'))
            site = sites.find(self._deco('VirtualNetworkSite'))
            site.set('Location', location)
            site.set('name', vnet_name)
        except Exception as e:
            log.error('_create_vnet_to_xml: %s' % e)
            raise
        print xmltree.tostring(tree)
        return xmltree.tostring(tree)

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_resource_lock_internal']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self._resource_lock_internal = None


class AzureVM:

    def __init__(
            self,
            config,
            node_index,
            cloud_service=None,
            storage_account=None,
            subscription=None,
            flavor=None,
            image=None,
            node_name=None,
            host_name=None,
            image_userdata=None):
        self._do_cleanup = False
        self._config = config
        self._node_index = node_index
        self._cloud_service = cloud_service
        self._storage_account = storage_account
        self._subscription = subscription
        self._flavor = flavor
        self._image = image
        self._node_name = node_name
        self._host_name = host_name
        self._image_userdata = image_userdata

        # figure out what sub, cloud serv, stor acct to use if not specified
        if not self._cloud_service:
            parent = self._config._parent
            (i_subscription, i_cloud_service, i_vm_in_cs, i_storage_account,
             i_vm_in_sa) = self._config._vm_flat_to_resources(self._node_index)
            self._subscription = parent._subscriptions[i_subscription]
            self._cloud_service = \
                self._subscription._cloud_services[i_cloud_service]
            self._storage_account = \
                self._subscription._storage_accounts[i_storage_account]

        self._qualified_name = '{0}_vm{1:04d}'.format(
            self._config._base_name, self._node_index)
        self._public_ip_internal = None
        self._ssh_port = None
        self._os_virtual_hard_disk = None
        self._os_vhd_name = None
        self._created = False
        self._paused = False

    def _delete(self):
        try:
            if len(self._cloud_service._instances) == 1:
                # we are the last node in this cloud service,
                # so delete deployment
                self._cloud_service._delete_deployment()
            else:
                result = self._cloud_service._subscription._sms.delete_role(
                    service_name=self._cloud_service._name,
                    deployment_name=self._cloud_service._name,
                    role_name=self._qualified_name)
                self._subscription._wait_result(result)
            self._storage_account._delete_vhd(self._os_vhd_name)
            with self._cloud_service._resource_lock:
                del(self._cloud_service._instances[self._qualified_name])
            # TODO - delete/reset our self?
        except Exception as e:
            log.error('error deleting vm %s: %s' % (self._qualified_name, e))
            raise

    def _start(self):
        try:
            self._create_network_config()

            (self._os_virtual_hard_disk, _) = \
                self._storage_account._create_vhd(self._node_name, self._image)
            virtual_network_name = \
                self._config._parent._subscriptions[0]._vnet._name

            sms = self._cloud_service._subscription._sms
            if self._cloud_service._instances:
                # add a node to existing deployement
                result = sms.add_role(
                    service_name=self._cloud_service._name,
                    deployment_name=self._cloud_service._name,
                    role_name=self._qualified_name,
                    system_config=self._system_config,
                    network_config=self._network_config,
                    os_virtual_hard_disk=self._os_virtual_hard_disk,
                    role_size=self._flavor,
                    role_type='PersistentVMRole')
                self._subscription._wait_result(result)
            else:
                # create the deployment and first node
                result = sms.create_virtual_machine_deployment(
                    service_name=self._cloud_service._name,
                    deployment_name=self._cloud_service._name,
                    deployment_slot='production',
                    label=self._node_name,
                    role_name=self._qualified_name,
                    system_config=self._system_config,
                    network_config=self._network_config,
                    os_virtual_hard_disk=self._os_virtual_hard_disk,
                    role_size=self._flavor,
                    role_type='PersistentVMRole',
                    virtual_network_name=virtual_network_name)
                self._subscription._wait_result(result)
        except Exception as e:
            if str(e) == 'Conflict (Conflict)':
                log.debug('virtual machine %s already exists.'
                          % self._qualified_name)
            else:
                log.error('error creating vm %s: %s'
                          % (self._qualified_name, e))
            raise
        self._created = True
        with self._cloud_service._resource_lock:
            self._cloud_service._instances[self._qualified_name] = self

        # need to find the disk name for the OS disk attached to this vm.
        # this involves redundant work that should be dealt with if it's
        # a problem.
        self._subscription._find_os_disks()

    def pause(self, instance_id, keep_provisioned=True):
        """shuts down the instance without destroying it.

        The AbstractCloudProvider class uses 'stop' to refer to destroying
        a VM, so use 'pause' to mean powering it down while leaving it
        allocated.

        :param str instance_id: instance identifier

        :return: None
        """
        try:
            node_info = self._instances.get(instance_id)
            if node_info is None:
                raise Exception(
                    "could not get state for instance %s" % instance_id)
            if node_info['PAUSED']:
                log.debug("node %s is already paused" % instance_id)
                return
            if node_info.get('FIRST'):
                # TODO - determine if any special logic needed for this
                # node
                pass
            node_info['PAUSED'] = True
            post_shutdown_action = 'Stopped' if keep_provisioned else \
                'StoppedDeallocated'
            result = self._sms[0].shutdown_role(
                service_name=self._cloud_service_name,
                deployment_name=self._deployment_name,
                role_name=instance_id,
                post_shutdown_action=post_shutdown_action)
            self._subscription._wait_result(result)
        except Exception as e:
            log.error("error pausing instance %s: %s" % (instance_id, e))
            raise
        log.debug('paused instance(instance_id=%s)' % instance_id)

    def restart(self, instance_id):
        """restarts a paused instance.

        :param str instance_id: instance identifier

        :return: None
        """
        try:
            node_info = self._instances.get(instance_id)
            if node_info is None:
                raise Exception(
                    "could not get state for instance %s" % instance_id)
            if not node_info['PAUSED']:
                log.debug(
                    'node %s is not paused, can\'t restart' % instance_id)
                return
            if node_info.get('FIRST'):
                # TODO - determine if any special logic needed for this
                # node
                pass
            node_info['PAUSED'] = False
            result = self._sms[0].start_role(
                service_name=self._cloud_service_name,
                deployment_name=self._deployment_name,
                role_name=instance_id)
            self._subscription._wait_result(result)
        except Exception as e:
            log.error('error restarting instance %s: %s' %
                      (instance_id, e))
            raise
        log.debug('restarted instance(instance_id=%s)' % instance_id)

    def _create_network_config(self):
        # Create linux configuration
        self._system_config = azure.servicemanagement.LinuxConfigurationSet(
            self._node_name,
            self._config._username,
            None,
            disable_ssh_password_authentication=True)
        ssh_config = azure.servicemanagement.SSH()
        ssh_config.public_keys = azure.servicemanagement.PublicKeys()
        authorized_keys_path = u'/home/%s/.ssh/authorized_keys' \
                               % self._config._username
        ssh_config.public_keys.public_keys.append(
            azure.servicemanagement.PublicKey(
                path=authorized_keys_path,
                fingerprint=self._subscription._fingerprint))
        self._system_config.ssh = ssh_config

        # Create network configuration
        self._network_config = azure.servicemanagement.ConfigurationSet()
        self._network_config.configuration_set_type = 'NetworkConfiguration'
        if self._config._use_public_ips:
            public_ip = azure.servicemanagement.PublicIP(
                u'pip-%s' % self._node_name)
            # allowed range is 4-30 mins
            public_ip.idle_timeout_in_minutes = 30
            public_ips = azure.servicemanagement.PublicIPs()
            public_ips.public_ips.append(public_ip)
            self._network_config.public_ips = public_ips
            self._ssh_port = SSH_PORT
        else:
            # create endpoints for ssh (22). Map to an offset + instance
            # index + port # for the public side
            self._ssh_port = PORT_MAP_OFFSET + self._node_index + SSH_PORT

        endpoints = azure.servicemanagement.ConfigurationSetInputEndpoints()
        endpoints.subnet_names = []
        endpoints.input_endpoints.append(
            azure.servicemanagement.ConfigurationSetInputEndpoint(
                name='TCP-%s' %
                self._ssh_port,
                protocol='TCP',
                port=self._ssh_port,
                local_port=SSH_PORT))
        self._network_config.input_endpoints = endpoints

    @property
    def _power_state(self):
        instances = self._cloud_service._deployment.role_instance_list
        for instance in instances:
            if instance.instance_name == self._qualified_name:
                # cache IP since it will be asked for soon
                if self._config._use_public_ips:
                    self._public_ip_internal = instance.public_ips[0].address
                else:
                    self._public_ip_internal = instance.instance_endpoints[
                        0].vip
                return instance.power_state
        raise Exception("could not get power_state for instance %s"
                        % self._qualified_name)

    @property
    def _public_ip(self):
        if not self._public_ip_internal:
            # not cached, so look it up
            instances = self._cloud_service._deployment.role_instance_list
            for instance in instances:
                if instance.instance_name == self._qualified_name:
                    if self._config._use_public_ips:
                        self._public_ip_internal = instance.public_ips[
                            0].address
                    else:
                        self._public_ip_internal = instance.instance_endpoints[
                            0].vip
                    return instance.power_state
            raise Exception("could not get public IP for instance %s"
                            % self._qualified_name)
        return self._public_ip_internal


class AzureCloudProvider(AbstractCloudProvider):

    """This implementation of
    :py:class:`elasticluster.providers.AbstractCloudProvider` uses the
    Azure Python interface connect to the Azure clouds and manage instances.

    An AzureCloudProvider owns a tree of Azure resources, rooted in one or
    more subscriptions and one or more storage accounts.
    """

    def __init__(self,
                 subscription_id,
                 certificate,
                 storage_path=None):
        """The constructor of AzureCloudProvider class is called only
        using keyword arguments.

        Usually these are configuration option of the corresponding
        `setup` section in the configuration file.
        """
        # Paramiko debug level
        # logging.getLogger('paramiko').setLevel(logging.DEBUG)
        # logging.basicConfig(level=logging.DEBUG)

        # for winpdb debugging
        # import rpdb2
        # rpdb2.start_embedded_debugger('food')

        # Ansible debug level
        import ansible
        import ansible.utils
        ansible.utils.VERBOSITY = 2

        # flag indicating resource creation failed - don't even
        # attempt node operations
        self._start_failed = None

        # this lock should never be held for long - only for changes to the
        # resource arrays this object owns, and queries about same.
        self._resource_lock_internal = None

        # resources
        self._subscriptions = []
        self._storage_accounts = []

        self._config = AzureGlobalConfig(
            self, subscription_id, certificate, storage_path)

    def start_instance(
            self,
            key_name,
            public_key_path,
            private_key_path,
            security_group,
            flavor,
            image,
            image_userdata,
            location=None,
            base_name=None,
            username=None,
            node_name=None,
            host_name=None,
            use_public_ips=None,
            wait_timeout=None,
            use_short_vm_names=None,
            n_cloud_services=None,
            n_storage_accounts=None,
            **kwargs):
        """Starts a new instance on the cloud using the given properties.
        Multiple instances might be started in different threads at the same
        time. The implementation should handle any problems regarding this
        itself.
        :return: str - instance id of the started instance
        """

        if self._start_failed:
            raise Exception('start_instance for node %s: failing due to'
                            ' previous errors.' % node_name)

        # locking is rudimentary at this point
        with self._resource_lock:
            # it'd be nice if elasticluster called something like
            # init_cluster() with all the args that will be the
            # same for every node created. But since it doesn't, handle that on
            # first start_instance call.
            if not self._config._setup_done:
                self._config.setup(
                    key_name,
                    public_key_path,
                    private_key_path,
                    security_group,
                    location,
                    base_name=base_name,
                    username=username,
                    use_public_ips=use_public_ips,
                    wait_timeout=wait_timeout,
                    use_short_vm_names=use_short_vm_names,
                    n_cloud_services=n_cloud_services,
                    n_storage_accounts=n_storage_accounts,
                    **kwargs)

            # absolute node index in cluster (0..n-1) determines what
            # subscription, cloud service, storage
            # account, etc. this VM will use. Cloud provider only tracks
            # how many instances have been created.
            index = self._n_instances
            if index == 0:
                self._create_global_reqs()
                if self._start_failed:
                    return None

            vm = AzureVM(
                self._config,
                index,
                flavor=flavor,
                image=image,
                node_name=node_name,
                host_name=host_name,
                image_userdata=image_userdata)

            vm._start()
            log.debug('started instance %s' % vm._qualified_name)
            return vm._qualified_name

    def stop_instance(self, instance_id):
        """Stops the instance gracefully.

        :param str instance_id: instance identifier

        :return: None
        """
        if self._start_failed:
            raise Exception('stop_instance for node %s: failing due to'
                            ' previous errors.' % instance_id)

        with self._resource_lock:
            try:
                vm = self._qualified_name_to_vm(instance_id)
                if not vm:
                    err = "stop_instance: can't find instance %s" % instance_id
                    log.error(err)
                    raise Exception(err)
                vm._delete()
                # note: self._n_instances is a derived property, doesn't need
                # to be updated
                if self._n_instances == 0:
                    log.debug('last instance deleted, destroying '
                              'global resources')
                    self._delete_global_reqs()
            except Exception as exc:
                log.error('error stopping instance %s: %s' %
                          (instance_id, exc))
                raise
        log.debug('stopped instance %s' % instance_id)

    def get_ips(self, instance_id):
        """Retrieves the private and public ip addresses for a given instance.
        Note: Azure normally provides access to vms from a shared load
        balancer IP and
        mapping of ssh ports on the vms. So by default, the Azure provider
        returns strings
        of the form 'ip:port'. However, 'stock' elasticluster and ansible
        don't support this,
        so _use_public_ips uses Azure PublicIPs to expose each vm on the
        internet with its own IP
        and using the standard SSH port.

        :return: list (IPs)
        """
        if self._start_failed:
            raise Exception('get_ips for node %s: failing due to'
                            ' previous errors.' % instance_id)

        ret = list()
        vm = self._qualified_name_to_vm(instance_id)
        if not vm:
            raise Exception("Can't find instance_id %s" % instance_id)
        if self._config._use_public_ips:
            ret.append(vm._public_ip)
        else:
            ret.append("%s:%s" % (vm._public_ip, vm._ssh_port))

        log.debug('get_ips (instance %s) returning %s' %
                  (instance_id, ', '.join(ret)))
        return ret

    def is_instance_running(self, instance_id):
        """Checks if the instance is up and running.

        :param str instance_id: instance identifier

        :return: bool - True if running, False otherwise
        """
        if self._start_failed:
            raise Exception('is_instance_running for node %s: failing due to'
                            ' previous errors.' % instance_id)

        vm = self._qualified_name_to_vm(instance_id)
        if not vm:
            raise Exception("Can't find instance_id %s" % instance_id)
        return vm._power_state == 'Started'

    # ------------------ add-on methods ---------------------------------
    # (not part of the base class, but useful extensions)

    # -------------------- private members ------------------------------

    @property
    def _resource_lock(self):
        if self._resource_lock_internal is None:
            self._resource_lock_internal = threading.Lock()
        return self._resource_lock_internal

    @property
    def _n_instances(self):
        ret = 0
        for subscription in self._subscriptions:
            ret = ret + subscription._n_instances
        return ret

    def _qualified_name_to_vm(self, qualified_name):
        i_sub, i_cs, i_vm_in_cs, _, _ = \
            self._config._vm_name_to_resources(qualified_name)
        cs = self._subscriptions[i_sub]._cloud_services[i_cs]
        # TODO can't lock use be confined to its class?
        with cs._resource_lock:
            return cs._instances[qualified_name]

    def _create_global_reqs(self):
        try:
            for index in range(self._config._n_cloud_services):
                i_sub = floor_div(index, self._config._cs_per_sub)
                sub = self._subscriptions[i_sub]
                acs = AzureCloudService(self._config, sub, index)
                acs._create()
                acs._add_certificate()
                sub._cloud_services.append(acs)

            for index in range(self._config._n_storage_accounts):
                i_sub = floor_div(index, self._config._sa_per_sub)
                sub = self._subscriptions[i_sub]
                asa = AzureStorageAccount(self._config, sub, index)
                asa._create()
                sub._storage_accounts.append(asa)

            # one vnet should be enough for ~ 1000 vms. Tie it to
            # the first subscription.
            sub = self._subscriptions[0]
            sub._vnet = AzureVNet(self._config, sub, 0)
            sub._vnet._create()
        except Exception as e:
            log.error('_create_global_reqs error: %s' % e)
            self._start_failed = True
            raise

    # tear down non-node-specific resources. Current default is to delete
    # everything; this may change.
    #
    def _delete_global_reqs(self):
        try:
            for sub in self._subscriptions:
                for cs in sub._cloud_services:
                    if cs._instances:
                        err = "cloud service %s can't be destroyed, it " \
                              "still has %i vms." % \
                              (cs._name, len(cs._instances))
                        log.error(err)
                        raise Exception(err)
                    cs._delete()

                for sa in self._storage_accounts:
                    sa._delete()

            self._subscriptions[0]._vnet._delete()

        except Exception as e:
            log.error('_delete_global_reqs error: %s' % e)
            raise

    # methods to support pickling

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_resource_lock_internal']
        return d

    def __setstate__(self, state):
        self.__dict__ = state
        self._resource_lock_internal = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-# 
# @(#)gc3pie_config.py
# 
# 
# Copyright (C) 2013-2014, GC3, University of Zurich. All rights reserved.
# 
# 
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

__docformat__ = 'reStructuredText'
__author__ = 'Antonio Messina <antonio.s.messina@gmail.com>'

from ConfigParser import RawConfigParser
import re
from StringIO import StringIO

from elasticluster import log

# NODELIST          NODES PARTITION       STATE CPUS    S:C:T MEMORY TMP_DISK WEIGHT FEATURES REASON
# compute[001-002]      2    cloud*        idle    1    1:1:1    491     5026      1   (null) none
slurm_sinfo_regexp =  re.compile('^(?P<hostnames>[^ \t]+)\s+'
                                 '(?P<num>[0-9]+)\s+'
                                 '(?P<partition>[^ \t]+)\s+'
                                 '(?P<state>[^ \t]+)\s+'
                                 '(?P<cpus>[0-9]+)\s+'
                                 '(?P<S>[0-9]+):(?P<C>[0-9]+):(?P<T>[0-9]+)\s+'
                                 '(?P<memory>[0-9]+)\s+'
                                 '(?P<tmp_disk>[0-9]+)\s+'
                                 '(?P<weight>[0-9]+)\s+'
                                 '(?P<features>[^ ]+)\s+'
                                 '(?P<reason>[^ \t]+)')

slurm_scontrol_maxtime_regexp = re.compile('.*\sMaxTime=(?P<MaxTime>[^ \t]+)\s+')

def inspect_slurm_cluster(ssh, node_information):
    (_in, _out, _err) = ssh.exec_command("sinfo -hNel")

    nodes = []
    for line in _out:
        match = slurm_sinfo_regexp.match(line)
        if match:
            num_nodes = int(match.group('num'))
            num_cores = int(match.group('cpus')) * num_nodes
            memory = int(match.group('memory')) * num_nodes
            memory_per_core = float(match.group('memory')) / num_cores
            nodes.append([num_nodes, num_cores, memory, memory_per_core])
        else:
            log.warning("Unable to parse output of sinfo: following line doesn't match node regexp: '%s'" % line.strip())
    node_information['num_nodes'] = sum(i[0] for i in nodes)
    node_information['max_cores'] = sum(i[1] for i in nodes)
    node_information['max_cores_per_job'] = max(i[1] for i in nodes)
    node_information['max_memory_per_core'] = max(i[2] for i in nodes)

    (_in, _out, _err) = ssh.exec_command("scontrol -o show part")
    # Assuming only one partition
    line = _out.read()
    match = slurm_scontrol_maxtime_regexp.match(line)
    node_information['max_walltime'] = '672hours'
    if match:
        maxtime = match.group('MaxTime')
        if maxtime != 'UNLIMITED':
            node_information['max_walltime'] = maxtime
    
    return node_information


def inspect_pbs_cluster(ssh):
    pass


sge_qhost_regexp = re.compile('(?P<hostname>[^\s]+)\s+'
                              '(?P<arch>[^\s]+)\s+'
                              '(?P<ncpus>[0-9]+)\s+'
                              '(?P<load>[^\s]+)\s+'
                              '(?P<memory>[0-9\.MGT]+)\s+')

# This function is took from GC3Pie, http://code.google.com/p/gc3pie/
# module gc3pie.gc3libs.utils
def to_bytes(s):
    """
    Convert string `s` to an integer number of bytes.  Suffixes like
    'KB', 'MB', 'GB' (up to 'YB'), with or without the trailing 'B',
    are allowed and properly accounted for.  Case is ignored in
    suffixes.

    Examples::

      >>> to_bytes('12')
      12
      >>> to_bytes('12B')
      12
      >>> to_bytes('12KB')
      12000
      >>> to_bytes('1G')
      1000000000

    Binary units 'KiB', 'MiB' etc. are also accepted:

      >>> to_bytes('1KiB')
      1024
      >>> to_bytes('1MiB')
      1048576

    """
    last = -1
    unit = s[last].lower()
    if unit.isdigit():
        # `s` is a integral number
        return int(s)
    if unit == 'b':
        # ignore the the 'b' or 'B' suffix
        last -= 1
        unit = s[last].lower()
    if unit == 'i':
        k = 1024
        last -= 1
        unit = s[last].lower()
    else:
        k = 1000
    # convert the substring of `s` that does not include the suffix
    if unit.isdigit():
        return int(s[0:(last+1)])
    if unit == 'k':
        return int(float(s[0:last])*k)
    if unit == 'm':
        return int(float(s[0:last])*k*k)
    if unit == 'g':
        return int(float(s[0:last])*k*k*k)
    if unit == 't':
        return int(float(s[0:last])*k*k*k*k)
    if unit == 'p':
        return int(float(s[0:last])*k*k*k*k*k)
    if unit == 'e':
        return int(float(s[0:last])*k*k*k*k*k*k)
    if unit == 'z':
        return int(float(s[0:last])*k*k*k*k*k*k*k)
    if unit == 'y':
        return int(float(s[0:last])*k*k*k*k*k*k*k*k)

def inspect_sge_cluster(ssh, node_information):
    (_in, _out, _err) = ssh.exec_command("qhost")
    nodes = []
    for line in _out:
        match = sge_qhost_regexp.match(line)
        if match:
            nodes.append((match.group('hostname'),
                         int(match.group('ncpus')),
                         to_bytes(match.group('memory'))))
    node_information['num_nodes'] = len(nodes)
    node_information['max_cores'] = sum(i[1] for i in nodes)
    node_information['max_cores_per_job'] = node_information['max_cores']
    node_information['max_memory_per_core'] = max(i[2] for i in nodes)
    # No easy way to see the maximum walltime for a SGE cluster. We
    # should run qstat -g c to list the queues, and then run qconf -sq
    # <queue> and look for s_rt and h_rt
    node_information['max_walltime'] = '672hours'

def inspect_node(node):
    """
    This function accept a `elasticluster.cluster.Node` class,
    connects to a node and tries to discover the kind of batch system
    installed, and some other information.
    """
    node_information = {}
    ssh = node.connect()
    if not ssh:
        log.error("Unable to connect to node %s", node.name)
        return

    (_in, _out, _err) = ssh.exec_command("(type >& /dev/null -a srun && echo slurm) \
                      || (type >& /dev/null -a qconf && echo sge) \
                      || (type >& /dev/null -a pbsnodes && echo pbs) \
                      || echo UNKNOWN")
    node_information['type'] = _out.read().strip()

    (_in, _out, _err) = ssh.exec_command("arch")
    node_information['architecture'] = _out.read().strip()

    if node_information['type'] == 'slurm':
        inspect_slurm_cluster(ssh, node_information)
    elif node_information['type'] == 'sge':
        inspect_sge_cluster(ssh, node_information)
    ssh.close()
    return node_information

def create_gc3pie_config_snippet(cluster):
    """
    Create a configuration file snippet to be used with GC3Pie.
    """
    auth_section = 'auth/elasticluster_%s' % cluster.name
    resource_section = 'resource/elasticluster_%s' % cluster.name

    cfg = RawConfigParser()
    cfg.add_section(auth_section)

    frontend_node = cluster.get_frontend_node()
    cfg.set(auth_section, 'type', 'ssh')
    cfg.set(auth_section, 'username', frontend_node.image_user)

    cluster_info = inspect_node(frontend_node)
    cfg.add_section(resource_section)
    cfg.set(resource_section, 'enabled', 'yes')
    cfg.set(resource_section, 'transport', 'ssh')
    cfg.set(resource_section, 'frontend', frontend_node.preferred_ip)
    if not cluster_info:
        log.error("Unable to gather enough information from the cluster. "
                  "Following informatino are only partial!")
        cluster_info = {'architecture': 'unknown',
                        'type': 'unknown',
                        'max_cores': -1,
                        'max_cores_per_job': -1,
                        'max_memory_per_core': -1,
                        'max_walltime': '672hours'}

    cfg.set(resource_section, 'type', cluster_info['type'])
    cfg.set(resource_section, 'architecture', cluster_info['architecture'])
    cfg.set(resource_section, 'max_cores', cluster_info.get('max_cores', 1))
    cfg.set(resource_section, 'max_cores_per_job', cluster_info.get('max_cores_per_job', 1))
    cfg.set(resource_section, 'max_memory_per_core', cluster_info.get('max_memory_per_core', '2GB'))
    cfg.set(resource_section, 'max_walltime', cluster_info.get('max_walltime', '672hours'))
    
    cfgstring = StringIO()
    cfg.write(cfgstring)

    return cfgstring.getvalue()
    
